unit AlignPnl;

{ TAlignPanel: a panel that aligns the controls on it automatically.
  Delphi 1 + 2, 3 not tested yet.
  Version 0.2 Jul-16-1997
  (C) 1997 Christoph R. Kirchner

Disclaimer:

  Users of this component must accept this disclaimer of  warranty:
    "This unit is supplied as is. The author disclaims all warranties,
    expressed or implied, including, without limitation, the warranties
    of merchantability and of fitness for any purpose.
    The author assumes no liability for damages, direct or
    consequential, which may result from the use of this component."

  This component is donated to the public as public domain.

  This component can be freely used and distributed in commercial and
  private environments.

  The source code may be freely distributed and used. The author
  accepts no responsibility for its use or misuse.

  If you do find this component handy and you feel guilty for using
  such a great product without paying someone - sorry :)

  Please forward any comments or suggestions to Christoph Kirchner at:
  ckirchner@geocities.com

PS:

  Maybe you can find an update of this component at
  'The Component Building Site':
  http://www.geocities.com/SiliconValley/Heights/7874/delphi.htm

  Thanks to Eugene I Levin for a bugfix.
}

interface

uses
{$IFDEF WIN32}
  Windows,
{$ELSE DEF WIN32}
  WinTypes, WinProcs, Menus,
{$ENDIF DEF WIN32}
  SysUtils, Classes, Graphics, Controls, Forms,
  Messages, StdCtrls, ExtCtrls, DsgnIntf;

type

{ Invisible WinControl: }

  TTransparentControl = class(TWinControl)
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure WMEraseBkgnd(var Message: TWmEraseBkgnd);
              message WM_ERASEBKGND;
  public
    constructor Create(AOwner: TComponent); override;
  end;

{ WinControl that has to stay on top of another panel: }

  TTopControl = class(TTransparentControl)
  public
    constructor Create(AOwner: TComponent); override;
  end;


{ TAlignPanel }

  TAlignPanelAlignChildren = (acDefault, acHorizontal, acVertical);
  { acDefault: If the width > heitht, align Controls horizontally
    acHorizontal: Align Controls horizontally
    acVertical: Align Controls vertically
  }

  TAlignPanel = class(TCustomPanel)
  private
    ButtonWidth: Integer;
    ButtonHeight: Integer;
    FSpace: Integer;
    FTopControlCount: Integer;
    FTopControls: array [0..3] of TTopControl;
    FAlignChildren: TAlignPanelAlignChildren;
    FHorizontal: Boolean;
    procedure AdjustSize(ALeft, ATop: Integer; var W: Integer; var H: Integer);
    procedure SetSpace(Value: Integer);
    procedure SetAlignChildren(Value: TAlignPanelAlignChildren);
  protected
    procedure Loaded; override;
    procedure AlignControls(AControl: TControl; var Rect: TRect); override;
    procedure RealignButtons;
  public
    constructor Create(AOwner: TComponent); override;
    function GetControlsList: TList;
    property Horizontal: Boolean read FHorizontal;
  published
    property Align;
    property BevelInner;
    property BevelOuter;
    property BevelWidth;
    property BorderWidth;
    property BorderStyle;
    property Enabled;
    property Ctl3D;
    property Font;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property Visible;
  { new properties: }
    property AlignChildren: TAlignPanelAlignChildren
       read FAlignChildren write SetAlignChildren default acDefault;
  { The space between the child-Controls (0 to 8 pixels): }
    property Space: Integer read FSpace write SetSpace default 0;
  end;


procedure Register;

implementation

const
  sRegisterOnPage = 'Samples';

{$IFDEF WIN32}
  {$R AlignPnl.d32} { 32-bit-AlignPnl.dcr is renamed to d32 }
{$ELSE DEF WIN32}
  {$R AlignPnl.d16}
{$ENDIF DEF WIN32}



const
  MinBtnSize: TPoint = (X: 20; Y: 18);
  MaxSpaceWidth = 8;
  DesignClickBorder = 8;



{ TTransparentControl ------------------------------------------------------ }

constructor TTransparentControl.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := [];
end;

procedure TTransparentControl.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  Params.ExStyle := Params.ExStyle or WS_EX_TRANSPARENT;
end;

procedure TTransparentControl.WMEraseBkgnd(var Message: TWmEraseBkgnd);
begin  { do not erase background: }
  Message.Result := 1; { we have done anything that was to do }
end;


{ TTopControl -------------------------------------------------------------- }

constructor TTopControl.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Align := alClient;
  Caption := '';
  TabStop := false;
end;


{ TAlignPanel -------------------------------------------------------------- }

constructor TAlignPanel.Create(AOwner: TComponent);
const
  TPAlign: array [0..3] of TAlign = (alLeft, alTop, alRight, alBottom);
var
  i: Integer;
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle - [csSetCaption] + [csOpaque];
  BevelOuter := bvNone;
  BevelInner := bvNone;
  Width := 241;
  Height := 25;
  ButtonWidth := 0;
  ButtonHeight := 0;
  FSpace := 0;
  FAlignChildren := acDefault;
  if (csDesigning in ComponentState) then
  begin
    FTopControlCount := 4;
    for i := 0 to FTopControlCount - 1 do
    begin
      FTopControls[i] := TTopControl.Create(self);
      FTopControls[i].Parent := self;
      FTopControls[i].Align := TPAlign[i];
    end;
    FTopControls[0].Width := DesignClickBorder;
    FTopControls[1].Height := DesignClickBorder;
    FTopControls[2].Width := DesignClickBorder;
    FTopControls[3].Height := DesignClickBorder;
  end
  else
    FTopControlCount := 0;
end;

procedure TAlignPanel.Loaded;
var
  i: Integer;
begin
  inherited Loaded;
  for i := 0 to FTopControlCount - 1 do
    FTopControls[i].BringToFront;
end;

function TAlignPanel.GetControlsList: TList;
var
  i: Integer;
  c: Integer;
  ControlLeft: Integer;
  ControlTop: Integer;
  ControlTabOrder: Integer;
  AControlTabOrder: Integer;
  Inserted: Boolean;
begin
  result := TList.Create;
  try
    for i := 0 to ControlCount - 1 do
    begin
      if ((Controls[i] is TWinControl) and
          (not (Controls[i] is TTopControl))) or
         (Controls[i] is TGraphicControl) then
      begin
        ControlLeft := TControl(Controls[i]).Left;
        ControlTop := TControl(Controls[i]).Top;
        if (Controls[i] is TWinControl) then
          ControlTabOrder := TWinControl(Controls[i]).TabOrder
        else
          ControlTabOrder := -1;
        Inserted := false;
        for c := 0 to result.Count - 1 do
        begin
          if (TControl(result[c]) is TWinControl) then
            AControlTabOrder := TWinControl(result[c]).TabOrder
          else
            AControlTabOrder := -1;
          if (Horizontal and
              ((ControlLeft < TControl(result[c]).Left) or
               ((ControlLeft = TControl(result[c]).Left) and
                (ControlTabOrder < AControlTabOrder)))
              ) or
             (not Horizontal and
              ((ControlTop < TControl(result[c]).Top) or
               ((ControlTop = TControl(result[c]).Top) and
                (ControlTabOrder < AControlTabOrder)))) then
          begin
            result.Insert(c, Controls[i]);
            Inserted := true;
            break;
          end;
        end;
        if not Inserted then
          result.Add(Controls[i]);
      end;
    end;
  except
    result.Free;
  end;
end;


procedure TAlignPanel.AdjustSize(ALeft, ATop: Integer;
  var W: Integer; var H: Integer);
var
  ControlsList: TList;
  MinW: Integer;
  MinH: Integer;
  i: Integer;
  X: Integer;
  Y: Integer;
  SumOfSpace: Integer;
  AdjustedButtonWidth: Integer;
  AdjustedButtonHeight: Integer;
  Temp: Integer;
  Remain: Integer;
  ControlTabOrder: Integer;
begin
  if (csLoading in ComponentState) then Exit;
  if (AlignChildren = acDefault) then
    FHorizontal := (H < W)
  else
    FHorizontal := (AlignChildren = acHorizontal);
  ControlsList := GetControlsList;
  try
    if (ControlsList.Count > 0) then
    begin
      ControlTabOrder := 0;
      for i := 0 to ControlsList.Count - 1 do
        if (TControl(ControlsList[i]) is TWinControl) then
        begin
          TWinControl(ControlsList[i]).TabOrder := ControlTabOrder;
          Inc(ControlTabOrder);
        end;
      SumOfSpace := Pred(ControlsList.Count) * FSpace;
      if Horizontal then
      begin
        MinW := (ControlsList.Count * MinBtnSize.X) + SumOfSpace;
        if (W < MinW) then
          W := MinW;
        if (H < MinBtnSize.Y) then
          H := MinBtnSize.Y;
        ButtonHeight := H;
        ButtonWidth := (W - SumOfSpace) div ControlsList.Count;
        Temp := ControlsList.Count * ButtonWidth + SumOfSpace;
      { if (Align = alNone) then W := Temp; }
        X := ALeft;
        Remain := W - Temp;
        Temp := ControlsList.Count div 2;
        for i := 0 to ControlsList.Count - 1 do
        begin
          AdjustedButtonWidth := ButtonWidth;
          if (Remain <> 0) then
          begin
            Dec(Temp, Remain);
            if (Temp < 0) then
            begin
              Inc(Temp, ControlsList.Count);
              Inc(AdjustedButtonWidth);
            end;
          end;
          TControl(ControlsList[i]).SetBounds(
            X, ATop, AdjustedButtonWidth, H);
          Inc(X, AdjustedButtonWidth + Space);
        end;
      end
      else
      begin
        MinH := (ControlsList.Count * MinBtnSize.Y) + SumOfSpace;
        if (H < MinH) then
          H := MinH;
        if (W < MinBtnSize.X) then
          W := MinBtnSize.X;
        ButtonWidth := W;
        ButtonHeight := (H - SumOfSpace) div ControlsList.Count;
        Temp := ControlsList.Count * ButtonHeight + SumOfSpace;
      { if (Align = alNone) then H := Temp; }
        Y := ATop;
        Remain := H - Temp;
        Temp := ControlsList.Count div 2;
        for i := 0 to ControlsList.Count - 1 do
        begin
          AdjustedButtonHeight := ButtonHeight;
          if (Remain <> 0) then
          begin
            Dec(Temp, Remain);
            if (Temp < 0) then
            begin
              Inc(Temp, ControlsList.Count);
              Inc(AdjustedButtonHeight);
            end;
          end;
          TControl(ControlsList[i]).SetBounds(
            ALeft, Y, W, AdjustedButtonHeight);
          Inc(Y, AdjustedButtonHeight + Space);
        end;
      end;
    end;
  finally
    ControlsList.Free;
  end;
  for i := 0 to FTopControlCount - 1 do
    FTopControls[i].BringToFront;
end;

procedure TAlignPanel.RealignButtons;
var
  W, H: Integer;
var
  BevelSize: Integer;
  DecSize: Integer;
  DeltaClient: Integer;
begin
  BevelSize := BorderWidth;
  if BevelOuter <> bvNone then Inc(BevelSize, BevelWidth);
  if BevelInner <> bvNone then Inc(BevelSize, BevelWidth);
  try
    DeltaClient := (Width - ClientWidth);
  except
    DeltaClient := 0;
  end;
  DecSize := BevelSize * 2 + DeltaClient;
  W := Width - DecSize;
  H := Height - DecSize;
  AdjustSize(BevelSize, BevelSize, W, H);
  if (W <> Width - DecSize) or (H <> Height - DecSize) then
  begin
    inherited SetBounds(Left, Top, W + DecSize, H + DecSize);
  end;
end;

procedure TAlignPanel.AlignControls(AControl: TControl; var Rect: TRect);
begin
  inherited AlignControls(AControl, Rect);
  with Rect do
    if not (Left >= Right) and not (Top >= Bottom) then
      RealignButtons;
end;

procedure TAlignPanel.SetSpace(Value: Integer);
begin
  if (Value > MaxSpaceWidth) then
    Value := MaxSpaceWidth
  else
    if (Value < 0) then
      Value := 0;
  if (Value <> FSpace) then
  begin
    FSpace := Value;
    if not (csLoading in ComponentState) then
      RealignButtons;
  end;
end;

procedure TAlignPanel.SetAlignChildren(Value: TAlignPanelAlignChildren);
begin
  if (Value <> FAlignChildren) then
  begin
    FAlignChildren := Value;
    if not (csLoading in ComponentState) then
      RealignButtons;
  end;
end;


{ Register ------------------------------------------------------------------ }

type

  TWinControlTabOrderProperty = class(TTabOrderProperty)
  public
    procedure SetValue(const Value: string); override;
  end;

procedure TWinControlTabOrderProperty.SetValue(const Value: string);
var
  WinControl: TWinControl;
  AlignPanel: TAlignPanel;
  i: Integer;
  WinControlTabOrder: TTabOrder;
begin
  inherited SetValue(Value);
  WinControl := TWinControl(GetComponent(0));
  if (WinControl.Parent <> nil) and
     (WinControl.Parent is TAlignPanel) then
    AlignPanel :=
      TAlignPanel(WinControl.Parent)
  else
  begin
    exit;
  end;
  WinControlTabOrder := WinControl.TabOrder;
  if AlignPanel.Horizontal then
  begin
    for i := 0 to AlignPanel.ControlCount - 1 do
    begin
      if (AlignPanel.Controls[i] <> WinControl) and
         (not (AlignPanel.Controls[i] is TTopControl)) and
         (AlignPanel.Controls[i] is TWinControl) then
        with TWinControl(AlignPanel.Controls[i]) do
          if (TabOrder > WinControl.TabOrder) then
          begin
            if (Left <= WinControl.Left) then
              WinControl.Left := Left - 1;
          end
          else
            if (Left >= WinControl.Left) then
              WinControl.Left := Left + 1;
      WinControl.TabOrder := WinControlTabOrder;
    end;
  end
  else
  begin
    for i := 0 to AlignPanel.ControlCount - 1 do
    begin
      if (AlignPanel.Controls[i] <> WinControl) and
         (not (AlignPanel.Controls[i] is TTopControl)) and
         (AlignPanel.Controls[i] is TWinControl) then
        with TWinControl(AlignPanel.Controls[i]) do
          if (TabOrder > WinControl.TabOrder) then
          begin
            if (Top <= WinControl.Top) then
              WinControl.Top := Top - 1;
          end
          else
            if (Top >= WinControl.Top) then
              WinControl.Top := Top + 1;
      WinControl.TabOrder := WinControlTabOrder;
    end;
  end;
  AlignPanel.RealignButtons;
end;


{$IFDEF WIN32}

{ TAlignPanelEditor selects all components in an AlignPanel on doubleklick.
  It does not work with Delphi 1 because I do not know any method to select
  a component at designtime. }

type
  TAlignPanelEditor = class(TComponentEditor)
    procedure ExecuteVerb(Index: Integer); override;
    function GetVerb(Index: Integer): string; override;
    function GetVerbCount: Integer; override;
  end;

procedure TAlignPanelEditor.ExecuteVerb(Index: Integer);
var
  ControlsList: TList;
  ComponentList: TComponentList;
  i: Integer;
begin
  if not (Designer is TFormDesigner) then
    exit;
  ControlsList := TAlignPanel(Component).GetControlsList;
  try
    if (ControlsList.Count > 0) then
    begin
      ComponentList := TComponentList.Create;
      try
        for i := 0 to ControlsList.Count - 1 do
          ComponentList.Add(TControl(ControlsList[i]));
        try
          TFormDesigner(Designer).SetSelections(ComponentList);
        except end;
      finally
        ComponentList.Free;
      end;
    end;
  finally
    ControlsList.Free;
  end;
end;

function TAlignPanelEditor.GetVerb(Index: Integer): string;
begin
  Result := 'Select all components';
end;

function TAlignPanelEditor.GetVerbCount: Integer;
begin
  Result := 1;
end;
{$ENDIF DEF WIN32}




procedure Register;
begin
  RegisterComponents(sRegisterOnPage, [TAlignPanel]);
  RegisterPropertyEditor(TypeInfo(TTabOrder),
                         TWinControl,
                         'TabOrder',
                         TWinControlTabOrderProperty);
{$IFDEF WIN32}
  RegisterComponentEditor(TAlignPanel, TAlignPanelEditor);
{$ENDIF DEF WIN32}
end;

end.
